import time
from unittest.mock import MagicMock

import pytest
import click
from click.testing import CliRunner

from nci_cidc_cli import api
from nci_cidc_cli import upload

from .util import ExceptionCatchingThread

JOB_ID = -1
JOB_ETAG = "abcd"
GCS_BUCKET = "upload-bucket"
URL_MAPPING = {
    "local_path1.fastq.gz": "gcs/path/1234/fastq/2019-09-04T18:59:45.224099",
    "local_path2.fastq.gz": "gcs/path/4321/fastq/2019-09-04T18:59:45.224099",
}
EXTRA_METADATA = {"lp1": "uuid1"}
GCS_FILE_MAP = {
    "gcs/path/1234/fastq/2019-09-04T18:59:45.224099": "uuid2",
    "gcs/path/4321/fastq/2019-09-04T18:59:45.224099": "uuid3",
}
OPTIONAL_FILES = []
UPLOAD_TOKEN = "test-upload-token"


class UploadMocks:
    def __init__(self, monkeypatch):
        self.gcloud_login = MagicMock()
        monkeypatch.setattr("nci_cidc_cli.gcloud.login", self.gcloud_login)

        self.api_initiate_upload = MagicMock()
        self.api_initiate_upload.return_value = api.UploadInfo(
            JOB_ID,
            JOB_ETAG,
            GCS_BUCKET,
            URL_MAPPING,
            EXTRA_METADATA,
            GCS_FILE_MAP,
            OPTIONAL_FILES,
            UPLOAD_TOKEN,
        )
        monkeypatch.setattr(api, "initiate_upload", self.api_initiate_upload)

        self.upload_succeeded = MagicMock()
        monkeypatch.setattr(api, "upload_succeeded", self.upload_succeeded)

        self.upload_failed = MagicMock()
        monkeypatch.setattr(api, "upload_failed", self.upload_failed)

        self._poll_for_upload_completion = MagicMock()
        monkeypatch.setattr(
            upload, "_poll_for_upload_completion", self._poll_for_upload_completion
        )

        self._open_file_mapping = MagicMock()
        monkeypatch.setattr(upload, "_open_file_mapping", self._open_file_mapping)

        self.insert_extra_metadata = MagicMock()
        monkeypatch.setattr(
            "nci_cidc_cli.api.insert_extra_metadata", self.insert_extra_metadata
        )

    def assert_expected_calls(self, failure=False):
        self.gcloud_login.assert_called_once()
        self.api_initiate_upload.assert_called_once()
        if failure:
            self.upload_failed.assert_called_once_with(
                JOB_ID, UPLOAD_TOKEN, JOB_ETAG, GCS_FILE_MAP
            )
        else:
            self.upload_succeeded.assert_called_once_with(
                JOB_ID, UPLOAD_TOKEN, JOB_ETAG, GCS_FILE_MAP
            )
            self._poll_for_upload_completion.assert_called_once_with(
                JOB_ID, UPLOAD_TOKEN
            )


def run_isolated_upload(runner: CliRunner):
    """Run a test upload inside an isolated filesystem"""
    with runner.isolated_filesystem():
        run_upload(runner)


def run_upload(runner: CliRunner):
    """Run a test upload"""
    files = ["wes.xlsx"] + list(URL_MAPPING.keys())
    for fname in files:
        with open(fname, "wb") as f:
            f.write(b"blah blah metadata")
    upload.run_upload("wes", "wes.xlsx")


def test_upload_success(runner: CliRunner, monkeypatch):
    """
    Check that a successful upload call follows the expected execution flow.
    """
    mocks = UploadMocks(monkeypatch)

    upload_success = MagicMock()
    upload_success.return_value = GCS_FILE_MAP
    monkeypatch.setattr(upload, "_gsutil_assay_upload", upload_success)

    # Run a successful upload.
    run_isolated_upload(runner)

    upload_success.assert_called_once()
    mocks.assert_expected_calls()


def test_upload_interrupt(runner: CliRunner, monkeypatch):
    """
    Check that a KeyboardInterrupt-ed upload call alerts the API that the job errored.
    """
    mocks = UploadMocks(monkeypatch)

    # Simulate a keyboard interrupt
    upload_failure = MagicMock()
    upload_failure.side_effect = KeyboardInterrupt
    monkeypatch.setattr(upload, "_gsutil_assay_upload", upload_failure)

    # Run an interrupted upload.
    with pytest.raises(KeyboardInterrupt):
        run_isolated_upload(runner)

    mocks.assert_expected_calls(failure=True)


def test_upload_exception(runner: CliRunner, monkeypatch):
    """
    Check that a failed upload call alerts the API that the job errored.
    """
    mocks = UploadMocks(monkeypatch)

    # Simulate an exception
    upload_failure = MagicMock()
    upload_failure.side_effect = Exception("bad upload")
    monkeypatch.setattr(upload, "_gsutil_assay_upload", upload_failure)

    with pytest.raises(Exception, match="bad upload"):
        run_isolated_upload(runner)

    mocks.assert_expected_calls(failure=True)


def test_upload_api_initiate_exception(runner: CliRunner, monkeypatch):
    """
    Check that a failed upload call alerts the API that the job errored.
    """
    mocks = UploadMocks(monkeypatch)

    # Simulate an exception
    initiate_failure = MagicMock()
    initiate_failure.side_effect = api.ApiError("bad upload")
    monkeypatch.setattr(api, "initiate_upload", initiate_failure)

    with pytest.raises(Exception, match="bad upload"):
        run_isolated_upload(runner)

    mocks.gcloud_login.assert_called_once()


def test_poll_for_upload_completion(monkeypatch):
    """
    Check that the _poll_for_upload loop follows retry directions,
    times out, succeeds, and fails as expected.
    """
    _timeout_check = 0

    def get_did_timeout(n_tries=5):
        global _timeout_check
        _timeout_check = 0

        def did_timeout():
            global _timeout_check
            if _timeout_check == n_tries:
                return True
            _timeout_check += 1
            return False

        return did_timeout

    click_echo = MagicMock()
    monkeypatch.setattr(click, "echo", click_echo)
    monkeypatch.setattr(click, "secho", click_echo)

    def stdout():
        print(click_echo.call_args_list)
        stdout = "\n".join(
            [args[0][0] for args in click_echo.call_args_list if len(args[0])]
        )
        return stdout

    job_id = 1

    # Simulate a retry with timeout
    retry_in = 2
    retry_upload = MagicMock()
    retry_upload.return_value = api.MergeStatus(None, None, retry_in)
    monkeypatch.setattr(api, "poll_upload_merge_status", retry_upload)
    sleep = MagicMock()
    monkeypatch.setattr(time, "sleep", sleep)
    upload._poll_for_upload_completion(
        job_id, UPLOAD_TOKEN, _did_timeout_test_impl=get_did_timeout(4)
    )
    retry_upload.assert_called_with(job_id, UPLOAD_TOKEN)
    sleep.assert_called_with(1)
    assert len(sleep.call_args_list) == retry_in
    assert "timed out" in stdout()

    click_echo.reset_mock()

    # Simulate a success
    completed = MagicMock()
    completed.return_value = api.MergeStatus("merge-completed", None, None)
    monkeypatch.setattr(api, "poll_upload_merge_status", completed)
    upload._poll_for_upload_completion(
        job_id, UPLOAD_TOKEN, _did_timeout_test_impl=get_did_timeout(1)
    )
    completed.assert_called_once_with(job_id, UPLOAD_TOKEN)
    assert "succeeded" in stdout()

    click_echo.reset_mock()

    # Simulate a success
    failed = MagicMock()
    failed.return_value = api.MergeStatus("upload-failed", "some error details", None)
    monkeypatch.setattr(api, "poll_upload_merge_status", failed)
    upload._poll_for_upload_completion(
        job_id, UPLOAD_TOKEN, _did_timeout_test_impl=get_did_timeout(1)
    )
    failed.assert_called_once_with(job_id, UPLOAD_TOKEN)
    failure_stdout = stdout()
    assert "failed" in failure_stdout
    assert "some error details" in failure_stdout


def test_simultaneous_uploads(runner: CliRunner, monkeypatch):
    """
    Check that two uploads can run simultaneously without the CLI encountering
    errors. This is a smoketest, not a guarantee that the CLI robustly supports
    concurrent uploads for different assays.
    """
    UploadMocks(monkeypatch)

    gsutil_command = MagicMock()
    gsutil_command.return_value = MagicMock("subprocess")
    gsutil_command.return_value.args = ["gsutil", "arg1", "arg2"]
    gsutil_command.return_value.poll = lambda: 0
    gsutil_command.return_value.returncode = 0
    gsutil_command.return_value.stderr = MagicMock("stderr")
    gsutil_command.return_value.stderr.readline = lambda: "gsutil progress"
    monkeypatch.setattr("subprocess.Popen", gsutil_command)

    monkeypatch.setattr("nci_cidc_cli.auth.get_id_token", lambda: "test-token")

    def do_upload():
        run_upload(runner)

    with runner.isolated_filesystem():
        t1 = ExceptionCatchingThread(do_upload)
        t2 = ExceptionCatchingThread(do_upload)
        t1.start(), t2.start()
        t1.join(), t2.join()


def test_handle_upload_exc():
    """Check that exceptions are processed correctly"""
    with pytest.raises(KeyboardInterrupt, match="Upload canceled"):
        upload._handle_upload_exc(KeyboardInterrupt())

    with pytest.raises(RuntimeError, match="failed.\nfoo"):
        upload._handle_upload_exc(RuntimeError("foo"))


def test_gsutil_assay_upload(monkeypatch):
    """Check that _gsutil_assay_upload waits for all file upload processes to complete"""
    num_procs = 13
    process_mocks = [MagicMock(name=f"mock #{n}") for n in range(num_procs)]

    def _start_procs(*args):
        """Mock the _start_procs function"""
        for proc in process_mocks:
            proc.start()
            yield proc

    def _wait_for_upload(procs, total, optional_files):
        """Mock the _wait_for_upload function"""
        assert total == num_procs
        for proc in procs:
            proc.stop()
        return []

    monkeypatch.setattr(upload, "_start_procs", _start_procs)
    monkeypatch.setattr(upload, "_wait_for_upload", _wait_for_upload)
    monkeypatch.setattr(
        upload, "_compose_file_mapping", lambda *args: ([0] * num_procs, [])
    )

    upload._gsutil_assay_upload(
        api.UploadInfo(
            JOB_ID,
            JOB_ETAG,
            GCS_BUCKET,
            URL_MAPPING,
            EXTRA_METADATA,
            GCS_FILE_MAP,
            OPTIONAL_FILES,
            UPLOAD_TOKEN,
        ),
        "",
    )

    # Ensure that _gsutil_assay_upload has started and waited for every process
    for proc in process_mocks:
        proc.start.assert_called()
        proc.stop.assert_called()


def test_compose_file_mapping(tmpdir, monkeypatch):
    xlsx = str(tmpdir.join("bar.xlsx"))

    failing_map = {str(tmpdir.join("foo.bar")): "foo.bar"}
    upload_job = api.UploadInfo(
        JOB_ID,
        JOB_ETAG,
        GCS_BUCKET,
        failing_map,
        EXTRA_METADATA,
        GCS_FILE_MAP,
        OPTIONAL_FILES,
        UPLOAD_TOKEN,
    )
    with pytest.raises(Exception, match="Could not locate"):
        upload._compose_file_mapping(upload_job, xlsx)

    # doesn't fail if passed as an optional_file
    upload_job = api.UploadInfo(
        JOB_ID,
        JOB_ETAG,
        GCS_BUCKET,
        failing_map,
        EXTRA_METADATA,
        GCS_FILE_MAP,
        OPTIONAL_FILES + list(failing_map.keys()),
        UPLOAD_TOKEN,
    )
    upload_file_map, gcs_file_map = upload._compose_file_mapping(upload_job, xlsx)
    assert gcs_file_map == ["foo.bar"]

    local_map = {
        tmpdir.join("local.path"): "test/gcs.1a",
        tmpdir.join("test.dir"): "test/gcs.1b",
    }
    input_map = {
        "gs://bucket/gcs.path": "test/gcs.2",
        "gs://bucket/[brackets]/subitem": "test/gcs.3",
    }

    # files must exist without care for content, so write garbage
    for k in local_map.keys():
        k.write("foo")

    # but must be {str: str}
    input_map.update({str(k): v for k, v in local_map.items()})
    input_map["test.dir"] = input_map.pop(str(tmpdir.join("test.dir")))

    upload_job = api.UploadInfo(
        JOB_ID,
        JOB_ETAG,
        GCS_BUCKET,
        input_map,
        EXTRA_METADATA,
        GCS_FILE_MAP,
        OPTIONAL_FILES,
        UPLOAD_TOKEN,
    )

    # mock local file check
    isfile = MagicMock()
    isfile.return_value = True
    monkeypatch.setattr("os.path.isfile", isfile)

    # mock gsutil ls file check by returning input
    ls_return = MagicMock()
    ls_return.returncode = 0
    ls_return.stdout = (
        "gs://bucket:\ngs://bucket/gcs.path\n\ngs://bucket/[brackets]/subitem".encode(
            "utf-8"
        )
    )
    ls_subprocess = MagicMock()
    ls_subprocess.return_value = ls_return
    monkeypatch.setattr("subprocess.run", ls_subprocess)

    output_map, skipping = upload._compose_file_mapping(upload_job, xlsx)
    assert len(skipping) == 0, skipping
    ls_subprocess.assert_called_once()

    for k, v in output_map:
        if "local.path" in k:
            assert v == f"gs://{GCS_BUCKET}/test/gcs.1a"
        elif "dir" in k:
            assert k == str(tmpdir.join("test.dir"))
            assert v == f"gs://{GCS_BUCKET}/test/gcs.1b"
        elif "gcs.path" in k:
            assert v == f"gs://{GCS_BUCKET}/test/gcs.2"
        elif "brackets" in k:
            assert k == "gs://bucket/?brackets]/subitem"
            assert v == f"gs://{GCS_BUCKET}/test/gcs.3"

    # now don't return one of them and see it fail
    ls_return.stdout = "gs://bucket:\ngs://bucket/gcs.path\n".encode("utf-8")
    ls_subprocess.return_value = ls_return
    monkeypatch.setattr("subprocess.run", ls_subprocess)

    with pytest.raises(Exception, match=r"gs://bucket/\[brackets\]/subitem"):
        output_map, skipping = upload._compose_file_mapping(upload_job, xlsx)



from click.testing import CliRunner

from nci_cidc_cli import cache


def test_cache_hit(runner: CliRunner):
    """Test storing and getting an object from the cache"""
    key = "foo"
    value1 = "bar"
    value2 = "baz"

    with runner.isolated_filesystem():
        cache.store(key, value1)
        assert cache.get(key) == value1

        cache.store(key, value2)
        assert cache.get(key) == value2


def test_cache_miss(runner: CliRunner):
    """Test that we can't get an object that doesn't exist in the cache."""
    assert cache.get("missing key") is None



"""Upload local files to CIDC's upload bucket"""
import os
import time
import subprocess
from contextlib import contextmanager
from datetime import datetime
from typing import BinaryIO, Dict, Generator, List, Optional, Tuple

import click

from . import api
from . import gcloud


def run_upload(
    upload_type: str,
    xlsx_path: str,
    is_analysis: bool = False,
    is_testing: bool = False,
):
    """
    Upload data.

    Orchestrator execution flow:
    1. Log in to gcloud. The CLI user must be authenticated with
       gcloud to be able to upload to GCS.
    2. Make an initiate_upload request to the API. The API adds a
       record to the database tracking that the CLI user started an
       upload job, grants the CLI user write permissions to the CIDC
       upload bucket in GCS, and returns information needed to
       carry out the gsutil upload (like a mapping from local file paths
       to GCS URIs).
    3. Carry out the gsutil upload using the returned upload info.
    4. If the gsutil upload fails, alert the api that the job failed.
       Else, if the upload succeeds, alert the api that the job was
       successful.
    """
    if not is_testing:
        # Log in to gcloud (required for gsutil to work)
        click.secho("> authenticating with gcloud", dim=True)
        gcloud.login()

    try:
        click.secho("> preparing upload job via the CIDC API", dim=True)
        # Read the .xlsx file and make the API call
        # that initiates the upload job and grants object-level GCS access.
        with open(xlsx_path, "rb") as xlsx_file:
            upload_info = api.initiate_upload(upload_type, xlsx_file, is_analysis)

    except (Exception, KeyboardInterrupt) as e:
        _handle_upload_exc(e)

    try:
        # Insert extra metadata for the upload, if any
        if upload_info.extra_metadata:
            click.secho(
                f"> pulling additional metadata from files staged for upload", dim=True
            )
            with _open_file_mapping(
                upload_info.extra_metadata, xlsx_path
            ) as open_files:
                api.insert_extra_metadata(upload_info.job_id, open_files)

        # Actually upload the assay data
        click.secho(f"> initiating GCS upload", dim=True)
        gcs_file_map = _gsutil_assay_upload(upload_info, xlsx_path)
    except (Exception, KeyboardInterrupt) as e:
        # we need to notify api of a failed upload
        api.upload_failed(
            upload_info.job_id,
            upload_info.token,
            upload_info.job_etag,
            upload_info.gcs_file_map,
        )
        _handle_upload_exc(e)
        # _handle_upload_exc should raise, but raise for good measure
        raise
    else:
        api.upload_succeeded(
            upload_info.job_id, upload_info.token, upload_info.job_etag, gcs_file_map
        )

    click.secho("> finalizing upload via the CIDC API", dim=True)
    _poll_for_upload_completion(upload_info.job_id, upload_info.token)


@contextmanager
def _open_file_mapping(extra_metadata: dict, base_path: str) -> Dict[str, BinaryIO]:
    """
    Given a dictionary mapping local paths to artifact uuids, return
    a dictionary mapping artifact uuids to open files.
    """
    base_dir = os.path.abspath(os.path.dirname(base_path))

    open_files = {}
    for source_path, uuid in extra_metadata.items():

        # if user wants us to get file from GCS
        # and we want it to be analysed for extra_md
        # we say we don't support it
        if source_path.startswith("gs://"):
            raise Exception(
                "File transfers from Google Cloud Storage are not supported for this assay type."
                f" Please download locally the files that you wish to upload ({source_path}),"
                " update the file paths in your metadata Excel file, and try again"
            )

        source_path = os.path.join(base_dir, source_path)
        open_files[uuid] = open(source_path, "rb")
    try:
        yield open_files
    except:
        raise
    finally:
        for f in open_files.values():
            f.close()


# This is how `gsutil` warns user when s/he sends large files.
# We don't want them to see that or enable `parallel_composite_upload_threshold`
# because composite files are problematic to download - see:
# https://cloud.google.com/storage/docs/gsutil/commands/cp#parallel-composite-uploads
_IGNORED_WARN_LINES = set(
    map(
        str.strip,
        """==> NOTE: You are uploading one or more large file(s), which would run
significantly faster if you enable parallel composite uploads. This
feature can be enabled by editing the
"parallel_composite_upload_threshold" value in your .boto
configuration file. However, note that if you do this large files will
be uploaded as `composite objects
<https://cloud.google.com/storage/docs/composite-objects>`_,which
means that any user who downloads such objects will need to have a
compiled crcmod installed (see "gsutil help crcmod"). This is because
without a compiled crcmod, computing checksums on composite objects is
so slow that gsutil disables downloads of composite objects.""".split(
            "\n"
        ),
    )
)


def _start_procs(src_dst_pairs: list) -> Generator[subprocess.Popen, None, None]:
    """
    Starts multiple "gsutil cp" subprocesses.

    src_dst_pairs: a list of tuples (local file path, target GCS path)

    Yields subprocess.Popen objects
    """
    procs = []

    src_dst_stack = list(src_dst_pairs)

    while True:
        try:
            src, dst = src_dst_stack.pop()
        except IndexError:
            return

        # Construct the upload command
        gsutil_args = ["gsutil", "cp", src, dst]

        try:
            # Run the upload command
            p = subprocess.Popen(
                gsutil_args,
                universal_newlines=True,
                bufsize=1,  # line buffered so we can read output line by line
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            procs.append(p)
            yield p

        except OSError as e:
            # stopping already created processes
            for p in procs:
                p.kill()
            _handle_upload_exc(e)

        except Exception as e:
            # stopping already created processes
            for p in procs:
                p.kill()
            _handle_upload_exc(e)


def _wait_for_upload(
    procs: list, total: int, optional_files: List[str]
) -> Optional[str]:
    """
    Waits for all subprocesses and click.echos their stderr streams.
    Returns Optional[str] - an error message if an error has occurred during any if uploads
    """

    # First we account all already successfully finished procs
    finished = set([i for i, p in enumerate(procs) if p.poll() == 0])

    error = None

    # GCS upload errors are generally spread across two lines.
    # Since we consume stderr one line at a time will polling upload processes,
    # we need to save the previous stderr line for each process in order
    # to reconstruct a full GCS upload error.
    prev_errlines = {}
    while len(finished) != len(procs) and not error:
        for i, p in enumerate(procs):
            if i in finished:
                continue

            # start building user feedback for this process
            message = f"[{len(finished)}/{total} done] "
            message += click.style(f"(file {i + 1}) ", fg="bright_blue")

            # read stderr for this process
            errline = p.stderr.readline()

            if p.poll() is not None:
                p.stderr.close()
                finished.add(i)

                if p.returncode != 0:
                    message += click.style("!!! upload error !!! ", fg="red", bold=True)
                    message += p.args[-2]
                    click.echo(message)
                    # Reconstruct multiline GCS error message
                    error = f"{prev_errlines.get(i, '')}{errline}"
                    break

            # skipping "large file" warnings
            if errline.strip() in _IGNORED_WARN_LINES:
                continue

            if (
                errline
                and len(errline) > 2  # skip '* ' spinner lines
                and errline.split("]", 1)[0].endswith(
                    "1 files"
                )  # include gsutil output with upload progress
            ):
                message += errline.split("]", 1)[1].rstrip()
                message += f" {p.args[-2]}"
                click.echo(message)
            else:
                # This might be the first line of a multiline error message,
                # so save it.
                prev_errlines[i] = errline

    return error


# default from `gsutil -m` but maybe better to load from env
MAX_GSUTIL_PARALLEL_PROCESS = 12


def _gsutil_assay_upload(upload_info: api.UploadInfo, xlsx: str) -> Dict[str, str]:
    """
    Upload local assay data to GCS using gsutil.
    Return modified GCS file map with missing files removed
    """

    upload_pairs, skipping = _compose_file_mapping(upload_info, xlsx)
    file_count = len(upload_pairs)
    for s in skipping:
        upload_info.gcs_file_map.pop(s, "")

    proc_iter = _start_procs(upload_pairs)
    procs = []
    all_uploads_have_run = False
    while not all_uploads_have_run:

        # Here we start with just 1 parallel process and gradually
        # increase that to MAX_GSUTIL_PARALLEL_PROCESS doubling the number every time.
        try:
            current_count = len(procs) or 1  # 1 is for starters
            how_many_to_add = min(current_count, MAX_GSUTIL_PARALLEL_PROCESS)
            for _ in range(how_many_to_add):
                procs.append(next(proc_iter))
        except StopIteration:
            all_uploads_have_run = True

        err = _wait_for_upload(procs, file_count, upload_info.optional_files)

        if err:
            for p in procs:

                if p.poll() != 0:

                    # stopping all other processes
                    for other_p in procs:
                        if p != other_p:
                            other_p.kill()

                    click.echo(
                        f"\nGCS upload failed on {p.args[-2]} with the following message:\n"
                    )
                    click.secho(f"{err}", fg="red")

                    raise click.Abort()

    click.echo(
        f"[{file_count}/{file_count} done] All files uploaded to GCS and staged for ingestion."
    )

    return upload_info.gcs_file_map


def _check_for_gs_files(
    gs_uris_to_check: Dict[str, Dict[str, str]],
    optional_files: List[str],
    target_bucket: str,
):
    """Smart checking of gs:// URIs to ensure that files exist"""
    res, missing_required_files, missing_optional_files = [], [], []

    # separate by bucket, to do single ls per bucket
    # then check in the return for all the files
    for bucket, gs_file_mapping in gs_uris_to_check.items():
        sub = subprocess.run(
            ["gsutil", "ls", "-r", f"gs://{bucket}"], capture_output=True
        )
        if sub.returncode != 0:
            click.secho(
                f"Error getting {bucket} to check files: {sub.stderr.decode('utf-8'), }",
                fg="red",
                bold=True,
            )
            raise click.Abort()
        # didn't fail

        # return in the format
        # folder:
        # gs://bucket/[file]
        # gs://bucket/[file...]
        #
        # [folder...]:
        file_list = set()
        for f in sub.stdout.decode("utf-8").split("\n"):
            f = f.strip()
            if f and not f.endswith(":"):
                file_list.add(f)

        for gs_source_path, gcs_uri in gs_file_mapping.items():
            if gs_source_path not in file_list:
                if gs_source_path in optional_files:
                    missing_optional_files.append(gcs_uri)
                    continue
                else:
                    missing_required_files.append(gs_source_path)
            else:
                # gsutil treats brackets in a gs-uri as a character set
                # see https://cloud.google.com/storage/docs/gsutil/addlhelp/WildcardNames#other-wildcard-characters
                # this replaces opening with a generic wildcard, which can only map to only a single bucket
                res.append(
                    [
                        gs_source_path.replace("[", "?"),
                        f"gs://{target_bucket}/{gcs_uri}",
                    ]
                )

    return res, missing_required_files, missing_optional_files


def _compose_file_mapping(
    upload_info: api.UploadInfo, xlsx: str
) -> Tuple[Dict[str, str], List[str]]:
    """
    Returns a list of (source_path, target uri) pairs for all
    the files from the upload info relative to the `work dir`
    that is xlsx file locaction. If s source_path is a GCS uri,
    it will return it w/o change.
    """
    res = []
    missing_optional_files = []
    missing_required_files = []
    xlsx_dir = os.path.abspath(os.path.dirname(xlsx))

    gs_uris_to_check = {}
    for source_path, gcs_uri in upload_info.url_mapping.items():

        # if we're not copying from GCS to GCS, then
        # resolve local path against .xslx file dir
        if not source_path.startswith("gs://"):
            source_path = os.path.join(xlsx_dir, source_path)

            if not os.path.isfile(source_path):
                if any(source_path.endswith(f) for f in upload_info.optional_files):
                    missing_optional_files.append(gcs_uri)
                    continue
                else:
                    missing_required_files.append(source_path)
            res.append([source_path, f"gs://{upload_info.gcs_bucket}/{gcs_uri}"])

        else:
            # separate by bucket, to do single ls per bucket
            # gsutil treats brackets in a gs-uri as a character set
            # see https://cloud.google.com/storage/docs/gsutil/addlhelp/WildcardNames#other-wildcard-characters
            # this replaces opening with a generic wildcard, which can only map to only a single bucket
            bucket = source_path[5:].split("/")[0].replace("[", "?")
            if bucket not in gs_uris_to_check:
                gs_uris_to_check[bucket] = {}
            gs_uris_to_check[bucket][source_path] = gcs_uri

    gs_res, missing_required_gs_files, missing_optional_gs_files = _check_for_gs_files(
        gs_uris_to_check, upload_info.optional_files, upload_info.gcs_bucket
    )

    res.extend(gs_res)
    missing_required_files.extend(missing_required_gs_files)
    missing_optional_files.extend(missing_optional_gs_files)

    if missing_required_files:
        raise Exception(
            f'Could not locate the following required files:\n{", ".join(missing_required_files)}'
        )

    return res, missing_optional_files


def _poll_for_upload_completion(
    job_id: int, job_token: str, timeout: int = 600, _did_timeout_test_impl=None
):
    """Repeatedly check if upload finalization either failed or succeed"""
    cutoff = datetime.now().timestamp() + timeout

    did_timeout = _did_timeout_test_impl or (
        lambda: datetime.now().timestamp() >= cutoff
    )

    debug_info_message = f"Please include this info in your inquiry: (job_id={job_id})"

    while not did_timeout():
        status = api.poll_upload_merge_status(job_id, job_token)
        if status.retry_in:
            # Loop in one second increments, checking
            # for a timeout on each iteration.
            for _ in range(status.retry_in - 1):
                if did_timeout():
                    break
                click.echo(".", nl=False)
                time.sleep(1)
        elif status.status:
            if "merge-completed" == status.status:
                click.echo(click.style("✓", fg="green", bold=True))
                click.echo(
                    "Upload succeeded. Visit the CIDC Portal "
                    "file browser to view your upload."
                )
            else:
                if status.status_details:
                    click.echo("Upload failed with the following message:")
                    click.echo()
                    click.secho(status.status_details, fg="red", bold=True)
                    click.echo()
                else:
                    click.echo("Upload failed. ", nl=False)
                click.echo(
                    "Please contact a CIDC administrator "
                    "(NCICIDCAdmin@mail.nih.gov) if you need assistance."
                )
                click.echo(debug_info_message)
            return
        else:
            # we should never reach this code block
            raise

    click.secho("!!!", fg="yellow", bold=True)
    click.echo(
        "Upload timed out. Please contact a CIDC administrator "
        "(NCICIDCAdmin@mail.nih.gov) for assistance."
    )
    click.echo(debug_info_message)


def _handle_upload_exc(e: Exception):
    """Handle an exception thrown during an upload attempt."""
    if isinstance(e, KeyboardInterrupt):
        raise KeyboardInterrupt(f"Upload canceled.")
    raise type(e)(f"Upload failed.\n{e}") from e



import pandas as pd
from typing import Dict, List, Set

from .core import (
    DownloadableFiles,
    get_clinical_downloadable_files,
    get_misc_data_files,
    get_trial_if_exists,
    get_shipments,
    Session,
    TrialMetadata,
    UploadJobs,
)

# copied from cidc_schemas.prism.constants
# Windows doesn't handle current jsonschemas RefResolver configuration
SUPPORTED_ANALYSES: List[str] = [
    "atacseq_analysis",
    "cytof_analysis",
    "rna_level1_analysis",
    "tcr_analysis",
    "wes_analysis",
    "wes_tumor_only_analysis",
]
SUPPORTED_ASSAYS: List[str] = [
    "atacseq_fastq",
    "wes_fastq",
    "wes_bam",
    "olink",
    "cytof",
    "ihc",
    "elisa",
    "rna_fastq",
    "rna_bam",
    "mif",
    "tcr_adaptive",
    "tcr_fastq",
    "hande",
    "nanostring",
    "clinical_data",
    "misc_data",
    "ctdna",
    "microbiome",
    "mibi",
]

SUPPORTED_ASSAYS_AND_ANALYSES: Set[str] = {
    # everything once, without file type specifiers
    a.split("_fastq")[0].split("_bam")[0].split("_adaptive")[0]
    for a in SUPPORTED_ASSAYS + SUPPORTED_ANALYSES
} | {"wes_analysis_old", "wes_tumor_only_analysis_old"}


pd.set_option("display.max_columns", None)
pd.set_option("display.max_colwidth", None)
pd.set_option("display.max_rows", None)


def _describe_olink(metadata_json: dict) -> pd.DataFrame:
    ret = pd.DataFrame(columns=["batch_id", "file", "cimac_id"])
    for batch in metadata_json.get("assays", {}).get("olink", {"batches": []})[
        "batches"
    ]:
        if "combined" not in batch:
            for record in batch["records"]:
                ret = ret.append(
                    pd.DataFrame(
                        [
                            {
                                "batch_id": batch["batch_id"],
                                "file": record["files"]["assay_npx"]["object_url"],
                                "cimac_id": sample,
                            }
                            for sample in record["files"]["assay_npx"]["samples"]
                        ]
                    )
                )
        else:
            ret = ret.append(
                pd.DataFrame(
                    [
                        {
                            "batch_id": batch["batch_id"],
                            "file": "combined",
                            "cimac_id": sample,
                        }
                        for sample in batch["combined"]["npx_file"]["samples"]
                    ]
                )
            )

    return ret.reset_index(drop=True)


def _describe_elisa(metadata_json: dict) -> pd.DataFrame:
    # elisa is special case
    ret = pd.DataFrame(columns=["assay_run_id", "cimac_id"])
    for batch in metadata_json.get("assays", {}).get("elisa", []):
        ret = ret.append(
            pd.DataFrame(
                [
                    {"cimac_id": sample, "assay_run_id": batch["assay_run_id"]}
                    for sample in batch["assay_xlsx"]["samples"]
                ]
            )
        )
    return ret.reset_index(drop=True)


def _describe_nanostring(metadata_json: dict) -> pd.DataFrame:
    # nanostring is special case
    ret = pd.DataFrame(columns=["batch_id", "run_id", "cimac_id"])
    for batch in metadata_json.get("assays", {}).get("nanostring", []):
        for run in batch["runs"]:
            ret = ret.append(
                pd.DataFrame(
                    [
                        {
                            "batch_id": batch["batch_id"],
                            "run_id": run["run_id"],
                            "cimac_id": sample["cimac_id"],
                        }
                        for sample in run["samples"]
                    ]
                )
            )
    return ret.reset_index(drop=True)


def _describe_rna_analysis(metadata_json: dict) -> pd.DataFrame:
    return pd.DataFrame(
        [
            {"cimac_id": record["cimac_id"]}
            for record in metadata_json.get("analysis", {}).get(
                "rna_analysis", {"level_1": []}
            )["level_1"]
        ]
    )


def _describe_cytof_analysis(
    metadata_json: dict,
) -> pd.DataFrame:
    ret = pd.DataFrame(columns=["batch_id", "cimac_id"])
    for batch in metadata_json.get("assays", {}).get("cytof", []):
        if "astrolabe_analysis" in batch:
            ret = ret.append(
                pd.DataFrame(
                    [
                        {
                            "batch_id": batch["batch_id"],
                            "cimac_id": record["cimac_id"],
                        }
                        for record in batch["records"]
                        if "output_files" in record
                    ]
                )
            )
    return ret.reset_index(drop=True)


def _describe_wes_analysis(metadata_json: dict, just_old: bool = False) -> pd.DataFrame:
    ret = pd.DataFrame(columns=["run_id", "tumor_cimac_id", "normal_cimac_id"])

    for subkey in ["wes_analysis", "wes_analysis_old"]:
        if just_old and "old" not in subkey:
            continue

        ret = ret.append(
            pd.DataFrame(
                [
                    {
                        "run_id": run["run_id"],
                        "tumor_cimac_id": run["tumor"]["cimac_id"],
                        "normal_cimac_id": run["normal"]["cimac_id"],
                    }
                    for run in metadata_json.get("analysis", {}).get(
                        subkey, {"pair_runs": []}
                    )["pair_runs"]
                ]
            )
        )
    return ret.reset_index(drop=True)


def _describe_wes_tumor_only_analysis(
    metadata_json: dict, just_old: bool = False
) -> pd.DataFrame:
    ret = pd.DataFrame(columns=["cimac_id"])

    for subkey in ["wes_tumor_only_analysis", "wes_tumor_only_analysis_old"]:
        if just_old and "old" not in subkey:
            continue

        ret = ret.append(
            pd.DataFrame(
                [
                    {"cimac_id": run["tumor"]["cimac_id"]}
                    for run in metadata_json.get("analysis", {}).get(
                        subkey, {"runs": []}
                    )["runs"]
                ]
            )
        )
    return ret.reset_index(drop=True)


def _describe_batched(metadata_json: dict, assay_or_analysis: str) -> pd.DataFrame:
    ret = pd.DataFrame(columns=["batch_id", "cimac_id"])
    for batch in metadata_json.get(
        "analysis" if "analysis" in assay_or_analysis else "assays", {}
    ).get(assay_or_analysis, {"batches": []})["batches"]:
        ret = ret.append(
            pd.DataFrame(
                [
                    {
                        "batch_id": batch["batch_id"],
                        "cimac_id": record["cimac_id"],
                    }
                    for record in batch["records"]
                ]
            )
        )
    return ret.reset_index(drop=True)


def _describe_generic(metadata_json: dict, assay_or_analysis: str) -> pd.DataFrame:
    ret = pd.DataFrame(columns=["batch_id", "cimac_id"])
    for batch_num, batch in enumerate(
        metadata_json.get(
            "analysis" if "analysis" in assay_or_analysis else "assays", {}
        ).get(assay_or_analysis, [])
    ):
        ret = ret.append(
            pd.DataFrame(
                [
                    {
                        "batch_id": batch.get("batch_id", str(batch_num)),
                        "cimac_id": record["cimac_id"],
                    }
                    for record in batch["records"]
                ]
            )
        )
    return ret.reset_index(drop=True)


def list_data_cimac_ids(trial_id: str, assay_or_analysis: str) -> None:
    """
    Prints a table listing all samples for the given assay/analysis and trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to investigate
    assay_or_analysis: str
        the name of the assay / analysis to investigate
        must be in SUPPORTED_ASSAYS_AND_ANALYSES
    """
    if assay_or_analysis not in SUPPORTED_ASSAYS_AND_ANALYSES:
        print("Assay / analysis not supported:", assay_or_analysis)
        return

    elif assay_or_analysis == "clinical_data":
        list_clinical(trial_id)
        return
    elif assay_or_analysis == "misc_data":
        list_misc_data(trial_id)
        return

    with Session.begin() as session:
        trial: TrialMetadata = get_trial_if_exists(trial_id, session=session)
        cimac_ids: pd.DataFrame

        if assay_or_analysis == "olink":
            cimac_ids: pd.DataFrame = _describe_olink(metadata_json=trial.metadata_json)
        elif assay_or_analysis == "elisa":
            cimac_ids: pd.DataFrame = _describe_elisa(metadata_json=trial.metadata_json)
        elif assay_or_analysis == "nanostring":
            cimac_ids: pd.DataFrame = _describe_nanostring(
                metadata_json=trial.metadata_json
            )

        elif assay_or_analysis == "rna_level1_analysis":
            cimac_ids: pd.DataFrame = _describe_rna_analysis(
                metadata_json=trial.metadata_json
            )

        elif assay_or_analysis == "cytof_analysis":
            cimac_ids: pd.DataFrame = _describe_cytof_analysis(
                metadata_json=trial.metadata_json
            )

        elif assay_or_analysis in ["wes_analysis", "wes_analysis_old"]:
            cimac_ids: pd.DataFrame = _describe_wes_analysis(
                metadata_json=trial.metadata_json,
                just_old="old" in assay_or_analysis,
            )

        elif assay_or_analysis in [
            "wes_tumor_only_analysis",
            "wes_tumor_only_analysis_old",
        ]:
            cimac_ids: pd.DataFrame = _describe_wes_tumor_only_analysis(
                metadata_json=trial.metadata_json,
                just_old="old" in assay_or_analysis,
            )

        elif assay_or_analysis == "tcr_analysis":
            cimac_ids: pd.DataFrame = _describe_batched(
                metadata_json=trial.metadata_json, assay_or_analysis=assay_or_analysis
            )

        else:
            cimac_ids: pd.DataFrame = _describe_generic(
                metadata_json=trial.metadata_json, assay_or_analysis=assay_or_analysis
            )

        # business print
        print(cimac_ids)


def list_clinical(trial_id: str) -> None:
    """
    Prints a table describing all clinical files for the given trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to investigate
    """
    with Session.begin() as session:
        trial: TrialMetadata = get_trial_if_exists(trial_id, session=session)
        clinical_files: List[DownloadableFiles] = get_clinical_downloadable_files(
            trial_id, session=session
        )

        number_of_participants: Dict[str, int] = dict()
        comments: Dict[str, str] = dict()
        for record in trial.metadata_json.get("clinical_data", {}).get("records", []):
            object_url: str = record["clinical_file"]["object_url"]
            number_of_participants[object_url] = record["clinical_file"][
                "number_of_participants"
            ]
            if "comment" in record:
                comments[object_url] = record["comment"]

        print(
            pd.DataFrame(
                {
                    "object_url": f.object_url,
                    "filename": f.object_url.split("/clinical/")[1],
                    "num_participants": number_of_participants.get(f.object_url),
                    "created": f._created,
                    "comment": comments.get(f.object_url, pd.NA),
                }
                for f in clinical_files
            )
        )


def list_misc_data(trial_id: str) -> None:
    """
    Prints a table describing all misc_data files for the given trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to investigate
    """
    with Session.begin() as session:
        trial: TrialMetadata = get_trial_if_exists(trial_id, session=session)
        misc_data_files: List[DownloadableFiles] = get_misc_data_files(
            trial_id, session=session
        )

        descriptions: Dict[str, str] = dict()
        batch_numbers: Dict[str, int] = dict()
        for batch_idx, batch in enumerate(
            trial.metadata_json.get("assays", {}).get("misc_data", [])
        ):
            for file in batch["files"]:
                object_url = file["file"]["object_url"]
                batch_numbers[object_url] = batch_idx

                if "description" in file:
                    descriptions[object_url] = file["description"]

        print(
            pd.DataFrame(
                {
                    "batch_id": batch_numbers.get(f.object_url, pd.NA),
                    "object_url": f.object_url,
                    "filename": f.object_url.split("/misc_data/")[1],
                    "created": f._created,
                    "description": descriptions.get(f.object_url, pd.NA),
                }
                for f in misc_data_files
            )
        )


def list_shipments(trial_id: str) -> None:
    """
    Prints a table describing all shipments for the given trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to investigate
    """
    with Session.begin() as session:
        shipments: List[UploadJobs] = get_shipments(trial_id, session=session)
        print(
            pd.DataFrame(
                {
                    "upload_type": u.upload_type,
                    "manifest_id": u.metadata_patch["shipments"][0]["manifest_id"],
                    "num_samples": sum(
                        len(p["samples"]) for p in u.metadata_patch["participants"]
                    ),
                    "created": u._created,
                }
                for u in shipments
            )
        )



from datetime import datetime
from typing import Dict, List, Optional, Tuple

from .core import (
    DownloadableFiles,
    get_clinical_downloadable_files,
    get_shipments,
    get_trial_if_exists,
    Session,
    TrialMetadata,
    UploadJobs,
)
from .list import SUPPORTED_ASSAYS_AND_ANALYSES


def _get_all_object_urls(target: dict) -> List[str]:
    ret = []
    for key, value in target.items():
        if key == "object_url":
            ret.append(value)
        if isinstance(value, dict):
            ret.extend(_get_all_object_urls(value))
        if isinstance(value, list):
            for each in value:
                if isinstance(each, dict):
                    ret.extend(_get_all_object_urls(each))
                # no second-level lists
    return ret


def _remove_misc_data_from_blob(
    metadata_json: dict,
    trial_id: str,
    batch_id: str,
    filename: str = None,
) -> Tuple[Optional[dict], List[str]]:
    try:
        batch_id: int = int(batch_id)
    except ValueError:
        print(f"Error: for misc_data, batch_id must be an integer, not {batch_id}")
        return None, []

    if batch_id >= len(metadata_json.get("assays", {}).get("misc_data", [])):
        print(f"Cannot find misc_data batch {batch_id} for trial {trial_id}")
        return None, []

    object_urls_to_delete: List[str] = []
    if filename:
        filenames: List[str] = [
            file["file"]["object_url"].split("/misc_data/")[1]
            for file in metadata_json["assays"]["misc_data"][batch_id]["files"]
        ]
        if filename not in filenames:
            print(
                f"Cannot find file {filename} in batch {batch_id} for trial {trial_id}"
            )
        else:
            record: dict = metadata_json["assays"]["misc_data"][batch_id]["files"].pop(
                filenames.index(filename)
            )
            object_urls_to_delete.extend(_get_all_object_urls(record))

    else:
        batch: dict = metadata_json["assays"]["misc_data"].pop(batch_id)
        object_urls_to_delete.extend(_get_all_object_urls(batch))

    return metadata_json, object_urls_to_delete


def _remove_olink_from_blob(
    metadata_json: dict, trial_id: str, batch_id: str, filename: str = None
) -> Tuple[Optional[dict], List[str]]:
    batch_ids: List[str] = [
        batch["batch_id"]
        for batch in metadata_json.get("assays", {}).get("olink", {"batches": []})[
            "batches"
        ]
    ]
    if batch_id != "study" and batch_id not in batch_ids:
        print(f"Cannot find olink batch {batch_id} for trial {trial_id}")
        return None, []

    object_urls_to_delete: List[str] = []
    batch_idx: int = batch_ids.index(batch_id) if batch_id != "study" else None
    if filename:
        if filename == "combined":
            if "combined" not in metadata_json["assays"]["olink"]["batches"][batch_idx]:
                print(
                    f"Cannot find a combined file for olink batch {batch_id} for trial {trial_id}"
                )
                return None, []
            else:
                # otherwise just the one file
                target: dict = metadata_json["assays"]["olink"]["batches"][
                    batch_idx
                ].pop("combined")
                object_urls_to_delete.extend(_get_all_object_urls(target))
        else:
            object_urls: List[str] = [
                record["files"]["assay_npx"]["object_url"]
                for record in metadata_json["assays"]["olink"]["batches"][batch_idx][
                    "records"
                ]
            ]
            if filename not in object_urls:
                print(
                    f"Cannot find a file {filename} in olink batch {batch_id} for trial {trial_id}"
                )
                return None, []
            else:
                target: dict = metadata_json["assays"]["olink"]["batches"][batch_idx][
                    "records"
                ].pop(object_urls.index(filename))
                object_urls_to_delete.extend(_get_all_object_urls(target))

    elif batch_id == "study":
        # batch_idx is None
        target: dict = metadata_json["assays"]["olink"].pop("study")
        object_urls_to_delete.extend(_get_all_object_urls(target))

    else:
        target: dict = metadata_json["assays"]["olink"]["batches"].pop(batch_idx)
        object_urls_to_delete.extend(_get_all_object_urls(target))

    # also remove hanging structures
    target = dict()
    if filename:
        # if we didn't pop the batch but might have left it hanging
        # if we DID pop, batch_idx would either not exist or point to the following entry
        batch: dict = metadata_json["assays"]["olink"]["batches"][
            batch_idx
        ]  # for convenience
        if "combined" not in batch and not batch["records"]:
            # if no combined or any records, remove the whole batch
            target["batch"] = metadata_json["assays"]["olink"]["batches"].pop(batch_idx)
    if "study" not in metadata_json and not metadata_json["assays"]["olink"]["batches"]:
        # if no batches, remove the whole assay
        target["assay"] = metadata_json["assays"].pop("olink")
    if target:
        object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_elisa_from_blob(
    metadata_json: dict,
    trial_id: str,
    assay_run_id: str,
) -> Tuple[Optional[dict], List[str]]:
    assay_run_ids: List[str] = [
        assay_run["assay_run_id"]
        for assay_run in metadata_json.get("assays", {}).get("elisa", [])
    ]
    if assay_run_id not in assay_run_ids:
        print(f"Cannot find elisa batch {assay_run_id} for trial {trial_id}")
        return None, []

    object_urls_to_delete: List[str] = []
    assay_run_idx: int = assay_run_ids.index(assay_run_id)

    assay_run: dict = metadata_json["assays"]["elisa"].pop(assay_run_idx)
    object_urls_to_delete.extend(_get_all_object_urls(assay_run))

    # also remove hanging structures
    target = dict()
    if not metadata_json["assays"]["elisa"]:
        # if no batches, remove the whole assay
        target: dict = metadata_json["assays"].pop("elisa")
    if target:
        object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_nanostring_from_blob(
    metadata_json: dict,
    trial_id: str,
    batch_id: str,
    run_id: str = None,
) -> Tuple[Optional[dict], List[str]]:
    batch_ids: List[str] = [
        batch["batch_id"]
        for batch in metadata_json.get("assays", {}).get("nanostring", [])
    ]
    if batch_id not in batch_ids:
        print(f"Cannot find nanostring batch {batch_id} for trial {trial_id}")
        return None, []

    object_urls_to_delete: List[str] = []
    batch_idx: int = batch_ids.index(batch_id)
    if run_id:
        run_ids: List[str] = [
            run["run_id"]
            for run in metadata_json["assays"]["nanostring"][batch_idx]["runs"]
        ]
        if run_id not in run_ids:
            print(
                f"Cannot find a run {run_id} in nanostring batch {batch_id} for trial {trial_id}"
            )
            return None, []
        else:
            record: dict = metadata_json["assays"]["nanostring"][batch_idx]["runs"].pop(
                run_ids.index(run_id)
            )
            object_urls_to_delete.extend(_get_all_object_urls(record))
    else:
        batch: dict = metadata_json["assays"]["nanostring"].pop(batch_idx)
        object_urls_to_delete.extend(_get_all_object_urls(batch))

    # also remove hanging structures
    target = dict()
    if run_id:
        # if we didn't pop the batch but might have left it hanging
        # if we DID pop, batch_idx would either not exist or point to the following entry
        batch: dict = metadata_json["assays"]["nanostring"][
            batch_idx
        ]  # for convenience
        if not batch["runs"]:
            # if no combined or any runs, remove the whole batch
            target["batch"] = metadata_json["assays"]["nanostring"].pop(batch_idx)
    if not metadata_json["assays"]["nanostring"]:
        # if no batches, remove the whole assay
        target["assay"] = metadata_json["assays"].pop("nanostring")
    if target:
        object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_rna_analysis_from_blob(
    metadata_json: dict,
    trial_id: str,
    cimac_id: str,
) -> Tuple[Optional[dict], List[str]]:
    cimac_ids: List[str] = [
        record["cimac_id"]
        for record in metadata_json.get("analysis", {}).get(
            "rna_analysis", {"level_1": []}
        )["level_1"]
    ]
    if cimac_id not in cimac_ids:
        print(f"Cannot find RNA analysis for {cimac_id} for trial {trial_id}")
        return None, []

    object_urls_to_delete: List[str] = []
    record_idx: int = cimac_ids.index(cimac_id)

    record: dict = metadata_json["analysis"]["rna_analysis"]["level_1"].pop(record_idx)
    object_urls_to_delete.extend(_get_all_object_urls(record))

    # also remove hanging structures
    target = dict()
    if not metadata_json["analysis"]["rna_analysis"]["level_1"]:
        # if no batches, remove the whole assay
        target: dict = metadata_json["analysis"].pop("rna_analysis")
    if target:
        object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_cytof_analysis_from_blob(
    metadata_json: dict,
    trial_id: str,
    batch_id: str,
    cimac_id: str = None,
) -> Tuple[Optional[dict], List[str]]:
    batch_ids: List[str] = [
        batch["batch_id"]
        for batch in metadata_json.get("assays", {}).get("cytof", [])
        if "astrolabe_analysis" in batch
    ]
    if batch_id not in batch_ids:
        print(f"Cannot find cytof analysis batch {batch_id} for trial {trial_id}")
        return None, []
    else:
        batch_idx: int = batch_ids.index(batch_id)

    object_urls_to_delete: List[str] = []
    if cimac_id:
        cimac_ids: List[str] = [
            record["cimac_id"]
            for record in metadata_json["assays"]["cytof"][batch_idx]["records"]
            if "output_files" in record
        ]
        if cimac_id not in cimac_ids:
            print(
                f"Cannot find cytof analysis for sample {cimac_id} in batch {batch_id} for trial {trial_id}"
            )
            return None, []
        else:
            record_idx: int = cimac_ids.index(cimac_id)
            output_files: dict = metadata_json["assays"]["cytof"][batch_idx]["records"][
                record_idx
            ].pop("output_files")
            object_urls_to_delete.extend(_get_all_object_urls(output_files))

    else:
        batch: dict = {
            key: metadata_json["assays"]["cytof"][batch_idx].pop(key, {})
            for key in [
                "astrolabe_reports",
                "astrolabe_analysis",
                "control_files_analysis",
            ]
        }
        object_urls_to_delete.extend(_get_all_object_urls(batch))

        # remove analysis from all samples in this batch as well
        for record_idx, record in enumerate(
            metadata_json["assays"]["cytof"][batch_idx]["records"]
        ):
            if "output_files" in record:
                output_files: dict = metadata_json["assays"]["cytof"][batch_idx][
                    "records"
                ][record_idx].pop("output_files")
                object_urls_to_delete.extend(_get_all_object_urls(output_files))

    # there cannot be any hanging structure to remove
    return metadata_json, object_urls_to_delete


def _remove_wes_analysis_from_blob(
    metadata_json: dict,
    trial_id: str,
    run_id: str,
    just_old: bool = False,
) -> Tuple[Optional[dict], List[str]]:
    object_urls_to_delete: List[str] = []
    for subkey in ["wes_analysis", "wes_analysis_old"]:
        if just_old and "old" not in subkey:
            continue

        run_ids: List[str] = [
            run["run_id"]
            for run in metadata_json.get("analysis", {}).get(subkey, {"pair_runs": []})[
                "pair_runs"
            ]
        ]
        if run_id not in run_ids:
            # has to be missing from both to be an issue
            continue

        record_idx: int = run_ids.index(run_id)

        record: dict = metadata_json["analysis"][subkey]["pair_runs"].pop(record_idx)
        object_urls_to_delete.extend(_get_all_object_urls(record))

    if not len(object_urls_to_delete):
        print(
            f"Cannot find "
            + ("old " if just_old else "")
            + f"WES paired analysis for {run_id} for trial {trial_id}"
        )
        return None, []

    # also remove hanging structures
    for subkey in ["wes_analysis", "wes_analysis_old"]:
        target = dict()
        if (
            subkey in metadata_json["analysis"]
            and not metadata_json["analysis"][subkey]["pair_runs"]
        ):
            # if no batches, remove the whole assay
            target: dict = metadata_json["analysis"].pop(subkey)
        if target:
            object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_wes_tumor_only_analysis_from_blob(
    metadata_json: dict,
    trial_id: str,
    cimac_id: str,
    just_old: bool = False,
) -> Tuple[Optional[dict], List[str]]:
    object_urls_to_delete: List[str] = []
    for subkey in ["wes_tumor_only_analysis", "wes_tumor_only_analysis_old"]:
        if just_old and "old" not in subkey:
            continue

        cimac_ids: List[str] = [
            run["tumor"]["cimac_id"]
            for run in metadata_json.get("analysis", {}).get(subkey, {"runs": []})[
                "runs"
            ]
        ]
        if cimac_id not in cimac_ids:
            # has to be missing from both to be an issue
            continue

        record_idx: int = cimac_ids.index(cimac_id)

        record: dict = metadata_json["analysis"][subkey]["runs"].pop(record_idx)
        object_urls_to_delete.extend(_get_all_object_urls(record))

    if not len(object_urls_to_delete):
        print(
            f"Cannot find "
            + ("old " if just_old else "")
            + f"WES tumor-only analysis for {cimac_id} for trial {trial_id}"
        )
        return None, []

    # also remove hanging structures
    for subkey in ["wes_tumor_only_analysis", "wes_tumor_only_analysis_old"]:
        target = dict()
        if (
            subkey in metadata_json["analysis"]
            and not metadata_json["analysis"][subkey]["runs"]
        ):
            # if no batches, remove the whole assay
            target: dict = metadata_json["analysis"].pop(subkey)
        if target:
            object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_batched_assay_from_blob(
    metadata_json: dict,
    trial_id: str,
    assay_or_analysis: str,
    batch_id: str,
    cimac_id: str = None,
) -> Tuple[Optional[dict], List[str]]:
    subkey: str = "analysis" if "analysis" in assay_or_analysis else "assays"
    batch_ids: List[str] = [
        batch["batch_id"]
        for batch in metadata_json.get(subkey, {}).get(
            assay_or_analysis, {"batches": []}
        )["batches"]
    ]
    if batch_id not in batch_ids:
        print(f"Cannot find {assay_or_analysis} batch {batch_id} for trial {trial_id}")
        return None, []

    object_urls_to_delete: List[str] = []
    batch_idx: int = batch_ids.index(batch_id)
    if cimac_id:
        cimac_ids: List[str] = [
            record["cimac_id"]
            for record in metadata_json[subkey][assay_or_analysis]["batches"][
                batch_idx
            ]["records"]
        ]
        if cimac_id not in cimac_ids:
            print(
                f"Cannot find {assay_or_analysis} for sample {cimac_id} in batch {batch_id} for trial {trial_id}"
            )
            return None, []
        else:
            record: dict = metadata_json[subkey][assay_or_analysis]["batches"][
                batch_idx
            ]["records"].pop(cimac_ids.index(cimac_id))
            object_urls_to_delete.extend(_get_all_object_urls(record))

    else:
        batch: dict = metadata_json[subkey][assay_or_analysis]["batches"].pop(batch_idx)
        object_urls_to_delete.extend(_get_all_object_urls(batch))

    # also remove hanging structures
    target = dict()
    if cimac_id:
        # if we didn't pop the batch but might have left it hanging
        # if we DID pop, batch_idx would either not exist or point to the following entry
        batch: dict = metadata_json[subkey][assay_or_analysis]["batches"][
            batch_idx
        ]  # for convenience
        if not batch["records"]:
            # if no combined or any records, remove the whole batch
            target["batch"] = metadata_json[subkey][assay_or_analysis]["batches"].pop(
                batch_idx
            )
    if not metadata_json[subkey][assay_or_analysis]["batches"]:
        # if no batches, remove the whole assay/analysis
        target[subkey] = metadata_json[subkey].pop(assay_or_analysis)
    if target:
        object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def _remove_generic_assay_from_blob(
    metadata_json: dict,
    trial_id: str,
    assay_or_analysis: str,
    batch_id: str,
    cimac_id: str = None,
) -> Tuple[Optional[dict], List[str]]:
    subkey: str = "analysis" if "analysis" in assay_or_analysis else "assays"
    batch_ids: List[str] = [
        batch["batch_id"]
        for batch in metadata_json.get(subkey, {}).get(assay_or_analysis, [])
        if "batch_id" in batch
    ]
    if batch_id not in batch_ids:
        if (
            batch_id.isnumeric()
            and "." not in batch_id
            and int(batch_id) >= 0
            and int(batch_id)
            < len(metadata_json.get(subkey, {}).get(assay_or_analysis, []))
        ):
            # only convert if if possible and not in batch_ids
            batch_idx: int = int(batch_id)
        else:
            print(
                f"Cannot find {assay_or_analysis} batch {batch_id} for trial {trial_id}"
            )
            return None, []
    else:
        batch_idx: int = batch_ids.index(batch_id)

    object_urls_to_delete: List[str] = []
    if cimac_id:
        cimac_ids: List[str] = [
            record["cimac_id"]
            for record in metadata_json[subkey][assay_or_analysis][batch_idx]["records"]
        ]
        if cimac_id not in cimac_ids:
            print(
                f"Cannot find {assay_or_analysis} for sample {cimac_id} in batch {batch_id} for trial {trial_id}"
            )
            return None, []
        else:
            record: dict = metadata_json[subkey][assay_or_analysis][batch_idx][
                "records"
            ].pop(cimac_ids.index(cimac_id))
            object_urls_to_delete.extend(_get_all_object_urls(record))

    else:
        batch: dict = metadata_json[subkey][assay_or_analysis].pop(batch_idx)
        object_urls_to_delete.extend(_get_all_object_urls(batch))

    # also remove hanging structures
    target = dict()
    if cimac_id:
        # if we didn't pop the batch but might have left it hanging
        # if we DID pop, batch_idx would either not exist or point to the following entry
        batch: dict = metadata_json[subkey][assay_or_analysis][
            batch_idx
        ]  # for convenience
        if not batch["records"]:
            # if no combined or any records, remove the whole batch
            target["batch"] = metadata_json[subkey][assay_or_analysis].pop(batch_idx)
    if not metadata_json[subkey][assay_or_analysis]:
        # if no batches, remove the whole assay/analysis
        target[subkey] = metadata_json[subkey].pop(assay_or_analysis)
    if target:
        object_urls_to_delete.extend(_get_all_object_urls(target))

    return metadata_json, object_urls_to_delete


def remove_data(trial_id: str, assay_or_analysis: str, target_id: Tuple[str]) -> None:
    """
    Removes a data section completely, include its downloadle_files entries

    Parameters
    ----------
    trial_id: str
        the id of the trial to investigate
    assay_or_analysis: str
        the name of the assay / analysis to investigate
        must be in SUPPORTED_ASSAYS_AND_ANALYSES
    target_id: Tuple[str]
        the ids to find the data to remove
        it cannot go past where is divisible in the data
            eg if ASSAY_OR_ANALYSIS == "clinical_data", only `filename` is accepted
            eg if ASSAY_OR_ANALYSIS == "elisa", only `assay_run_id` is accepted
            eg if ASSAY_OR_ANALYSIS == "wes_analysis", only `run_id` is accepted
            eg if ASSAY_OR_ANALYSIS == "olink", `batch_id [file]` is assumed
                ie `batch_id` is required but `file` is optional
    """

    if assay_or_analysis not in SUPPORTED_ASSAYS_AND_ANALYSES:
        print("Assay / analysis not supported:", assay_or_analysis)
        return

    elif assay_or_analysis == "clinical_data":
        if len(target_id) == 1:
            remove_clinical(trial_id=trial_id, target_id=target_id[0])
        else:
            print(
                "Error: if ASSAY_OR_ANALYSIS == 'clinical_data', only `filename` is accepted"
            )
            print(
                "You can also directly use the command $ nci-cidc admin remove clinical TRIAL_ID FILE_NAME"
            )
        return

    with Session.begin() as session:
        trial: TrialMetadata = get_trial_if_exists(
            trial_id, with_for_update=True, session=session
        )

        metadata_json, object_urls_to_delete = None, []
        if assay_or_analysis == "misc_data":
            if len(target_id) and len(target_id) <= 2:
                metadata_json, object_urls_to_delete = _remove_misc_data_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    batch_id=target_id[0],
                    filename=None if len(target_id) == 1 else target_id[1],
                )
            else:
                print(
                    "Error: if ASSAY_OR_ANALYSIS == 'misc_data', only `batch_id [filename]` is accepted"
                )
        elif assay_or_analysis == "olink":
            if len(target_id) and len(target_id) <= 2:
                metadata_json, object_urls_to_delete = _remove_olink_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    batch_id=target_id[0],
                    filename=None if len(target_id) == 1 else target_id[1],
                )
            else:
                print(
                    "Error: if ASSAY_OR_ANALYSIS == 'olink', only `batch_id [file]` is accepted"
                )
        elif assay_or_analysis == "elisa":
            if len(target_id) == 1:
                metadata_json, object_urls_to_delete = _remove_elisa_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    assay_run_id=target_id[0],
                )
            else:
                print(
                    "Error: if ASSAY_OR_ANALYSIS == 'elisa', only `assay_run_id` is accepted"
                )
        elif assay_or_analysis == "nanostring":
            if len(target_id) and len(target_id) <= 2:
                metadata_json, object_urls_to_delete = _remove_nanostring_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    batch_id=target_id[0],
                    run_id=None if len(target_id) == 1 else target_id[1],
                )
            else:
                print(
                    "Error: if ASSAY_OR_ANALYSIS == 'olink', only `batch_id [file]` is accepted"
                )
        elif assay_or_analysis == "rna_level1_analysis":
            if len(target_id) == 1:
                metadata_json, object_urls_to_delete = _remove_rna_analysis_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    cimac_id=target_id[0],
                )
            else:
                print(
                    "Error: if ASSAY_OR_ANALYSIS == 'rna_level1_analysis', only `cimac_id` is accepted"
                )

        elif assay_or_analysis == "cytof_analysis":
            if len(target_id) <= 2:
                metadata_json, object_urls_to_delete = _remove_cytof_analysis_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    batch_id=target_id[0],
                    cimac_id=None if len(target_id) == 1 else target_id[1],
                )
            else:
                print(
                    "Error: if ASSAY_OR_ANALYSIS == 'cytof_analysis', only `batch_id [cimac_id]` is accepted"
                )
        elif assay_or_analysis in ["wes_analysis", "wes_analysis_old"]:
            if len(target_id) == 1:
                metadata_json, object_urls_to_delete = _remove_wes_analysis_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    run_id=target_id[0],
                    just_old="old" in assay_or_analysis,
                )
            else:
                print(
                    f"Error: if ASSAY_OR_ANALYSIS == '{assay_or_analysis}', only `run_id` is accepted"
                )
        elif assay_or_analysis in [
            "wes_tumor_only_analysis",
            "wes_tumor_only_analysis_old",
        ]:
            if len(target_id) == 1:
                (
                    metadata_json,
                    object_urls_to_delete,
                ) = _remove_wes_tumor_only_analysis_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    cimac_id=target_id[0],
                    just_old="old" in assay_or_analysis,
                )
            else:
                print(
                    f"Error: if ASSAY_OR_ANALYSIS == '{assay_or_analysis}', only `cimac_id` is accepted"
                )

        elif assay_or_analysis == "tcr_analysis":
            if len(target_id) and len(target_id) <= 2:
                metadata_json, object_urls_to_delete = _remove_batched_assay_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    assay_or_analysis=assay_or_analysis,
                    batch_id=target_id[0],
                    cimac_id=None if len(target_id) == 1 else target_id[1],
                )
            else:
                print(
                    f"Error: if ASSAY_OR_ANALYSIS == '{assay_or_analysis}', only `batch_id [cimac_id]` is accepted"
                )
        else:
            if len(target_id) and len(target_id) <= 2:
                metadata_json, object_urls_to_delete = _remove_generic_assay_from_blob(
                    metadata_json=trial.metadata_json,
                    trial_id=trial_id,
                    assay_or_analysis=assay_or_analysis,
                    batch_id=target_id[0],
                    cimac_id=None if len(target_id) == 1 else target_id[1],
                )
            else:
                print(
                    f"Error: if ASSAY_OR_ANALYSIS == '{assay_or_analysis}', only `batch_id [cimac_id]` is accepted"
                )

        if metadata_json is not None:
            # update the `trial_metadata`
            session.query(TrialMetadata).filter(
                TrialMetadata.trial_id == trial_id
            ).update(
                {
                    TrialMetadata.metadata_json: metadata_json,
                    TrialMetadata._updated: datetime.now(),
                }
            )

            # remove the `downloadable_files`
            num_deleted: int = (
                session.query(DownloadableFiles)
                .filter(DownloadableFiles.object_url.in_(object_urls_to_delete))
                .delete()
            )

            print(
                f"Updated trial {trial_id}, removing {assay_or_analysis} values {target_id}",
                f"along with {num_deleted} files",
            )


def remove_clinical(trial_id: str, target_id: str) -> None:
    """
    Removes a clinical file completely, include its downloadle_files entry

    Parameters
    ----------
    trial_id: str
        the id of the trial to affect
    target_id: str
        the object_url of the file to remove
        not including {trial_id}/clinical/
        special value * for all files for this trial
    """
    with Session.begin() as session:
        trial: TrialMetadata = get_trial_if_exists(
            trial_id, with_for_update=True, session=session
        )

        if target_id == "*":
            # get all of the clinical files
            targets: List[DownloadableFiles] = get_clinical_downloadable_files(
                trial_id, session=session
            )
        else:
            # get just the one(s) we're looking for
            targets: List[DownloadableFiles] = (
                session.query(DownloadableFiles)
                .filter(
                    DownloadableFiles.trial_id == trial_id,
                    DownloadableFiles.object_url == f"{trial_id}/clinical/{target_id}",
                )
                .all()
            )
        target_urls = [f.object_url for f in targets]

        if not len(targets):
            if target_id == "*":
                print(f"No clinical data files found for trial {trial_id}")
            else:
                print(f"Clinical data file {target_id} not found for trial {trial_id}")
            exit()

        # remove the file(s)
        trial.metadata_json["clinical_data"]["records"] = [
            r
            for r in trial.metadata_json.get("clinical_data", {}).get("records", [])
            if r["clinical_file"]["object_url"] not in target_urls
        ]

        # remove any hanging structure
        if not len(trial.metadata_json["clinical_data"]["records"]):
            trial.metadata_json.pop("clinical_data")

        # update the `trial_metadata`
        session.query(TrialMetadata).filter(TrialMetadata.trial_id == trial_id).update(
            {
                TrialMetadata.metadata_json: trial.metadata_json,
                TrialMetadata._updated: datetime.now(),
            }
        )

        # remove the `downloadable_files`
        for t in targets:
            session.delete(t)

        print(f"Updated trial {trial_id}, removing {len(targets)} clinical data files")


def _remove_samples_from_blob(
    metadata_json: dict,
    samples_to_remove: Dict[str, List[str]],
) -> dict:
    """
    Removes a participant completely if removing all of its samples

    Parameters
    ----------
    metadata_json: dict
        the full clinical trial metadata blob
    samples_to_remove: Dict[str, List[str]]
        map from `cimac_participant_id`s to a list of `cimac_id`s to remove
    """
    n: int = 0
    while n < len(metadata_json["participants"]):
        partic = metadata_json["participants"][n]
        cimac_partic_id = partic["cimac_participant_id"]
        if cimac_partic_id in samples_to_remove:
            if len(partic["samples"]) == len(samples_to_remove[cimac_partic_id]):
                metadata_json["participants"].pop(n)
                # don't increment because we popped
                continue

            else:
                to_remove = samples_to_remove[cimac_partic_id]
                metadata_json["participants"][n]["samples"] = [
                    s for s in partic["samples"] if s["cimac_id"] not in to_remove
                ]
        n += 1

    return metadata_json


def remove_shipment(trial_id: str, target_id: str) -> None:
    """
    Removes a shipment completely with all of its samples, including its upload_jobs entry
    Removes a participant completely if removing all of its samples

    Parameters
    ----------
    trial_id: str
        the id of the trial to affect
    target_id: str
        the manifest_id of the shipment to remove
    """
    with Session.begin() as session:
        trial: TrialMetadata = get_trial_if_exists(
            trial_id, with_for_update=True, session=session
        )

        # get all of the shipments
        shipments: List[UploadJobs] = get_shipments(trial_id, session=session)
        # find the one(s) we're looking for
        targets = [
            s
            for s in shipments
            if s.metadata_patch["shipments"][0]["manifest_id"] == target_id
        ]

        if not len(targets):
            print(f"Shipment {target_id} not found for trial {trial_id}")
            exit()

        # get all the associated samples
        samples_to_remove = {
            partic["cimac_participant_id"]: [
                sample["cimac_id"] for sample in partic["samples"]
            ]
            for upload in targets
            for partic in upload.metadata_patch.get("participants", [])
        }
        if not len(samples_to_remove):
            print(f"Shipment {target_id} for trial {trial_id} has no samples")
            exit()

        # remove the samples
        trial.metadata_json = _remove_samples_from_blob(
            metadata_json=trial.metadata_json,
            samples_to_remove=samples_to_remove,
        )
        # remove the shipment(s)
        trial.metadata_json["shipments"] = [
            s for s in trial.metadata_json["shipments"] if s["manifest_id"] != target_id
        ]

        # update the `trial_metadata`
        session.query(TrialMetadata).filter(TrialMetadata.trial_id == trial_id).update(
            {
                TrialMetadata.metadata_json: trial.metadata_json,
                TrialMetadata._updated: datetime.now(),
            }
        )

        # remove the `upload_jobs`
        for t in targets:
            session.delete(t)

        num_samples: int = sum(len(samples) for samples in samples_to_remove.values())
        num_partic: int = len(samples_to_remove)

        print(
            f"Updated trial {trial_id}, removing shipment {target_id}",
            f"along with {num_samples} samples across {num_partic} participants",
        )



"""Implements a client for the CIDC API running on Google App Engine"""
from functools import wraps
from collections import namedtuple
from typing import Optional, List, BinaryIO, NamedTuple, Dict, Callable

import click
import requests
import pyperclip

from . import auth, __version__
from .config import API_V2_URL, get_env


class ApiError(click.ClickException):
    pass


def _read_clipboard() -> str:
    """Read the current contents of the user's clipboard."""
    txt = pyperclip.paste()
    return txt


def _url(endpoint: str) -> str:
    """Append `endpoint` to the API's base URL"""
    endpoint = endpoint.lstrip("/")
    return f"{API_V2_URL}/{endpoint}"


def _error_message(response: requests.Response):
    try:
        message = response.json()["_error"]["message"]
        if response.status_code >= 500:
            message = f"API server error: {message}"
        if type(message) == dict and "errors" in message:
            message_lines = ["Multiple errors:"]
            message_lines.extend(
                [
                    click.style("* ", fg="red", bold=True) + message
                    for message in message["errors"]
                ]
            )
            return "\n".join(message_lines)
        else:
            return str(message)
    except:
        return f"API server encountered an error processing your request {response.status_code}"


_USER_AGENT = f"nci-cidc-cli/{__version__}"


def _with_auth(headers: dict = None, id_token: str = None) -> dict:
    """Add an id token to the given headers"""
    if not id_token:
        id_token = auth.get_id_token()
    return {
        "Authorization": f"Bearer {id_token}",
        # Also, include user agent with info about the CLI version
        "User-Agent": _USER_AGENT,
        **(headers or {}),
    }


def check_auth(id_token: str) -> Optional[str]:
    """Check if an id_token is valid by making a request to the base API URL."""
    response = requests.get(_url("/users/self"), headers=_with_auth(id_token=id_token))

    if response.status_code != 200:
        raise ApiError(_error_message(response))


def retry_with_reauth(api_request):
    """
    For a function `api_request` that returns a `Response` object, if that response
    has status code 403, prompt the user to enter a fresh ID token from the portal,
    and retry the request.
    """

    urls = {
        "prod": "https://cidc.nci.nih.gov/analyses/cli-instructions",
        "staging": "https://cidc-stage.nci.nih.gov/analyses/cli-instructions",
        "dev-int": "https://cidc-dev.nci.nih.gov/analyses/cli-instructions",
        "dev-cloudtwo": "https://cidctemp.crri.nci.nih.gov/analyses/cli-instructions",
        "stage-cloudtwo": "https://cidcstage.crri.nci.nih.gov/analyses/cli-instructions",
        "prod-cloudtwo": "https://cidcprod.crri.nci.nih.gov/analyses/cli-instructions"
    }
    TOKEN_URL = urls.get(get_env()) or urls["dev-int"]

    @wraps(api_request)
    def wrapped(*args, **kwargs):
        retry = True
        while retry:
            res = api_request(*args, **kwargs)
            # If the error isn't auth-related, break out of the retry loop.
            if res.status_code != 401:
                break

            # Check if the user got an error because they lack sufficient permissions.
            # If so, don't prompt them for another token, since they'll just get the same error again.
            error_message = _error_message(res)
            if "is not authorized to upload" in error_message:
                raise ApiError(error_message)

            # Prompt the user for a new ID token.
            while True:
                click.prompt(
                    (
                        "\nCIDC reauthentication required. Please copy a fresh identity token from the Portal "
                        f"to your clipboard at this URL:\n\n\t{TOKEN_URL}\n\n"
                        "Then, press 'enter' to paste your copied token below"
                    ),
                    default="enter",
                    show_default=False,
                )
                try:
                    id_token = _read_clipboard()
                except:
                    click.echo(
                        f"\n\nError: could not read token from clipboard.\n",
                        color="red",
                    )
                    retry = False
                    break
                click.echo(f"\n{id_token}\n")

                # Validate and cache the user's ID token. If the token is invalid,
                # inform the user, and re-prompt them for an identity token.
                try:
                    auth.validate_and_cache_token(id_token)
                    kwargs["headers"] = _with_auth(kwargs.get("headers"), id_token)
                    break
                except auth.AuthError:
                    click.echo("The token you entered is invalid.")

            if not retry:
                break

            # Rewind any file pointers to avoid sending empty files on retry
            if "files" in kwargs:
                for file in kwargs["files"].values():
                    file.seek(0)

        # Handle error responses
        if res.status_code != 200:
            raise ApiError(_error_message(res))

        return res

    return wrapped


class _RequestsWithReauth:
    def __init__(self):
        """Build a `request` instance with all methods wrapped in the `retry_with_reauth` decorator."""
        pass

    def __getattribute__(self, name):
        return retry_with_reauth(getattr(requests, name))


_requests_with_reauth = _RequestsWithReauth()


def list_assays() -> List[str]:
    """Get a list of all supported assays."""
    response = requests.get(_url("/info/assays"))
    assays = response.json()
    return assays


def list_analyses() -> List[str]:
    """Get a list of all supported analyses."""
    response = requests.get(_url("/info/analyses"))
    assays = response.json()
    return assays


class UploadInfo(NamedTuple):
    """Container for data we expect to get back from an initiate upload request"""

    job_id: int
    job_etag: str
    gcs_bucket: str
    url_mapping: dict
    extra_metadata: list
    gcs_file_map: dict
    optional_files: list
    token: str


def test_csms():
    """A simple API hit for a test of CSMS connection"""
    response = _requests_with_reauth.get(_url("/admin/test_csms"), headers=_with_auth())
    click.echo(response.json())


def refresh_trial(trial_id: str):
    _requests_with_reauth.patch(_url(f"/trial_metadata/{trial_id}/refresh_trial"), headers=_with_auth())


def initiate_upload(
    upload_type: str, xlsx_file: BinaryIO, is_analysis: bool = False
) -> UploadInfo:
    """
    Initiate an upload.

    Args:
        upload_type: the name of the API-supported assay
        xlsx_file: an open .xlsx file
        is_analysis: whether this is an analysis upload. If `False`,
                     then it's assumed to be an assay upload.

    Returns:
        UploadInfo: a mapping from local filepaths to GCS upload URIs,
        along with an upload job ID.
    """
    data = {"schema": upload_type}

    files = {"template": xlsx_file}

    endpoint = "upload_analysis" if is_analysis else "upload_assay"

    response = _requests_with_reauth.post(
        _url(f"/ingestion/{endpoint}"), headers=_with_auth(), data=data, files=files
    )

    try:
        upload_info = response.json()
        return UploadInfo(**upload_info)
    except:
        raise ApiError(
            "Cannot decode API response. You may need to update the CIDC CLI."
        )


def _update_upload_status(
    job_id: int, job_token: str, etag: str, status: str, gcs_file_map: Dict[str, str]
):
    """Update the status for an existing upload job"""
    url = _url(f"/upload_jobs/{job_id}")
    data = {"status": status, "gcs_file_map": gcs_file_map}

    if_match = {"If-Match": etag}
    response = _requests_with_reauth.patch(
        url, params={"token": job_token}, json=data, headers=_with_auth(if_match)
    )
    return response


def upload_succeeded(
    job_id: int, job_token: str, etag: str, gcs_file_map: Dict[str, str]
):
    """Tell the API that an upload job succeeded"""
    _update_upload_status(job_id, job_token, etag, "upload-completed", gcs_file_map)


def insert_extra_metadata(job_id: int, extra_metadata: Dict[str, BinaryIO]):
    """Insert extra metadata into the patch for the given job"""
    data = {"job_id": job_id}

    response = _requests_with_reauth.post(
        _url("/ingestion/extra-assay-metadata"),
        headers=_with_auth(),
        data=data,
        files=extra_metadata,
    )

    if response.ok:
        return response
    else:
        raise ApiError(_error_message(response))


def upload_failed(job_id: int, job_token: str, etag: str, gcs_file_map: Dict[str, str]):
    """Tell the API that an upload job failed"""
    _update_upload_status(job_id, job_token, etag, "upload-failed", gcs_file_map)


class MergeStatus(NamedTuple):
    status: Optional[str]
    status_details: Optional[str]
    retry_in: Optional[int]


def poll_upload_merge_status(job_id: int, job_token: str) -> MergeStatus:
    """Check the merge status of an upload job"""
    url = _url(f"/ingestion/poll_upload_merge_status/{job_id}")
    params = {"token": job_token}

    response = _requests_with_reauth.get(url, params=params, headers=_with_auth())

    merge_status = response.json()
    status = merge_status.get("status")
    status_details = merge_status.get("status_details")
    retry_in = merge_status.get("retry_in")

    if not (status or retry_in):
        raise ApiError("The server responded with an unexpected upload status message.")

    return MergeStatus(status, status_details, retry_in)
